/*		 ICP-PCI Kernel module free public release
 *
 *		       Copyright (C) 1996-99 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: osLx22.c,v 5.1.2.7 1999/09/22 11:47:12 matthias Exp $
 *
 */

/*
 * Includes:
 */

#include "icpciOS.h"	/* os specific stuff */
#include "icpciExt.h"	/* offical driver interface */
#include "icpciInt.h"	/* driver internals */
#include "icpciReg.h"	/* register definitions IC-PCI */

#include <linux/module.h>	/* init_module */

#ifdef MODULE
/* 
 * Command Line Options (insmod)
 *
 * Bus Master Access:
 * 'bma=disable', or no bma option -> disabled / else enabled
 *
 * Reserved Upper Memory Size
 * 'base=XXM', memory base in MBytes, only if autodetetion off
 * 'size=XXM', memory size in MBytes, only if autodetetion off
 *
 * Byte/Word/Doubleword Register Access Delay Time:
 * 'wait8=20', means 20 usecs or 20 loops (-DICP_LOOP_WAIT) delay
 * 'wait16=5', see above
 * 'wait32=0', see above
 *
 * Major Device Number assignement:
 * 'major0=x',  0       : dynamic assignement (default)
 * 'major1=x',  1 - 128 : explicitely define the number (1 - MAX_CHRDEV)
 *                > 128 : old style behavior (ICP_MAJOR_0|1)
 *		
 */

char *bma = "enable";
MODULE_PARM(bma, "s");

#ifdef ICP_AUTODETECT
extern unsigned long
bios_memory_end;	/* upper limit of physical memory */

extern unsigned long
icpci_memory_size;	/* memory used exclusively by the driver */
#else
static u_long base = 0;
static size_t size = 0;
MODULE_PARM(base, "1-979h");
MODULE_PARM(size, "1-979h");
#endif

#ifdef ICP_WAIT_LOOP
static int wait8 = 400;
static int wait16 = 200;
static int wait32 = 100;
MODULE_PARM(wait8, "1-10000i");
MODULE_PARM(wait16, "1-10000i");
MODULE_PARM(wait32, "1-10000i");
#else
static int wait8 = 4;
static int wait16 = 2;
static int wait32 = 1;
MODULE_PARM(wait8, "1-100i");
MODULE_PARM(wait16, "1-100i");
MODULE_PARM(wait32, "1-100i");
#endif

static u_int major0 = 0;
static u_int major1 = 0;
MODULE_PARM(major0, "1-" __MODULE_STRING(MAX_CHRDEV) "i");
MODULE_PARM(major1, "1-" __MODULE_STRING(MAX_CHRDEV) "i");
#endif

/*
 * Defines
 */

/*
 * Globals:
 */

extern struct 
icp_global_t    icp;	/* main driver structure */

char rcsid[] = "version 0.5.1 09/21/1999 (linux 2.2)";

static u_long used_intr = 0UL;	/* mark the used interrupts for the boards */
#define ICP_NUM_X86INTS	16	/* the x86 structure has up to 16 interrupts */

/*
 * Prototypes:
 */

static void
icp_os_intr (int, void*, struct pt_regs*);
static void
icp_os_timeout (u_long);
static void
icp_os_task_queue (void*);
static int
icp_os_open (struct inode*, struct file*);
static int
icp_os_release (struct inode*, struct file*);
static ssize_t
icp_os_read (struct file*, char*, size_t, loff_t*);
static ssize_t
icp_os_write (struct file*, const char*, size_t, loff_t*);
static int
icp_os_ioctl (struct inode*, struct file*, u_int, u_long); 
static int
icp_os_mmap (struct file*, struct vm_area_struct*);
static u_int
icp_os_poll (struct file*, struct poll_table_struct*);

static struct file_operations icp_fops = {
  NULL,		/* llseek */
  icp_os_read,	/* read */
  icp_os_write,	/* write */
  NULL,		/* readdir */
  icp_os_poll,	/* poll */
  icp_os_ioctl,	/* ioctl */
  icp_os_mmap,  /* mmap */
  icp_os_open,	/* open */
  NULL,		/* flush */
  icp_os_release,	/* release */
  NULL,		/* fsync */
  NULL,		/* fasync */
  NULL,		/* check_media_change */
  NULL,		/* revalidate */
  NULL		/* lock */
};

/*
 * delayed_memset_io
 */

static void
delayed_memset_io (void *data)
{
  struct icp_memset_t *memset = (struct icp_memset_t*)data;

  while (memset->size > BLOCK_CLEAR_SIZE)
    {
      memset_io (memset->addr, memset->value, BLOCK_CLEAR_SIZE);
      memset->addr += BLOCK_CLEAR_SIZE;
      memset->size -= BLOCK_CLEAR_SIZE;
      if (in_interrupt())
	/* in irq handler we have to add task queues */
	{
	  icp.osp[memset->board]->clr_t_queue.sync = 0;
	  icp.osp[memset->board]->clr_t_queue.routine = delayed_memset_io;
	  icp.osp[memset->board]->clr_t_queue.data =
	    (void*)&icp.osp[memset->board]->memset;
	  queue_task(&icp.osp[memset->board]->clr_t_queue, &tq_scheduler);
	}
      else
	schedule();
    }

  if (memset->size > 0)
    memset_io (memset->addr, memset->value, memset->size);

  if (in_interrupt())
    icp_wakeup (memset->board, CAMERA);
}

void*
icp_memset_io (int board, void *addr, char value, size_t size)
{
  void *dst_addr = addr;
  
  if (size > BLOCK_CLEAR_SIZE)
    {
      icp.osp[board]->memset.board = board;
      icp.osp[board]->memset.addr = addr;
      icp.osp[board]->memset.value = value;
      icp.osp[board]->memset.size = size;
      delayed_memset_io (&icp.osp[board]->memset);
    }
  else
    if (size > 0)
      memset_io (addr, value, size);
  
  return (dst_addr);
}

/*
 * icp_malloc
 */

void* 
icp_malloc (size_t size, int mode)
{
  switch (mode)
    {
    case UNPAGED:
      return kmalloc (size, GFP_KERNEL);
      break;
    case PAGED:
      return vmalloc (size);
      break;
    default:
      return NULL;
    }
  
  return NULL;
}

/*
 * icp_free
 */

void
icp_free(void *base, int mode)
{
  switch (mode)
    {
    case UNPAGED:
      return kfree (base);
      break;
    case PAGED:
      return vfree (base);
      break;
    }
}

/*
 * icp_add_timer
 */

int
icp_add_timer (int board, time_t timeout, int mode)
{
  struct timer_list *timer_ptr;

  switch (mode)
    {
    case CAMERA:
      timer_ptr = &icp.osp[board]->cam_timer;
      break;
    case IOPORT:
      timer_ptr = &icp.osp[board]->iop_timer;
      break;
    default:
      return -EINVAL;
    }

  if (!del_timer (timer_ptr))
    ICP_PRINTF ("icpci[%c]: timer is not queued.\n", 'W');

  timer_ptr->data = board;
  timer_ptr->function = icp_os_timeout;
  timer_ptr->expires = jiffies + timeout;
  add_timer (timer_ptr);
  
  return OK;
}

/*
 * icp_del_timer
 */

int
icp_del_timer (int board, int mode)
{
  struct timer_list *timer_ptr;

  switch (mode)
    {
    case CAMERA:
      timer_ptr = &icp.osp[board]->cam_timer;
      break;
    case IOPORT:
      timer_ptr = &icp.osp[board]->iop_timer;
      break;
    default:
      return -EINVAL;
    }
  
  if (!del_timer (timer_ptr))
    ICP_PRINTF ("icpci[%c]: timer is not queued.\n", 'W');
  
  return OK;
}

/*
 * icp_sleep
 */

int
icp_sleep (int board, time_t timeout, int mode)
{
  int error;

  switch (mode)
    {
    case CAMERA:
      if ((error = icp_add_timer (board, timeout, CAMERA)))
	return error;
      interruptible_sleep_on (&icp.osp[board]->cam_w_queue);
      if (current->signal.sig[0] & ~current->blocked.sig[0])
	return -ERESTARTSYS;
      break;
    case IOPORT:
      if ((error = icp_add_timer (board, timeout, IOPORT)))
	return error;
      interruptible_sleep_on (&icp.osp[board]->iop_w_queue);
      if (current->signal.sig[0] & ~current->blocked.sig[0])
	return -ERESTARTSYS;
      break;
    default:
      return -EINVAL;
    }

  return OK;
}

/*
 * icp_wakeup
 */

int
icp_wakeup (int board, int mode)
{
  int error;

  switch (mode)
    {
    case CAMERA:
      wake_up (&icp.osp[board]->cam_w_queue);
      if ((error = icp_del_timer (board, CAMERA)))
	return error;
      break;
    case IOPORT:
      wake_up (&icp.osp[board]->iop_w_queue);
      if ((error = icp_del_timer (board, IOPORT)))
	return error;
      break;
    default:
      return -EINVAL;
    }

  return OK;
}

/*
 * icp_marktq
 */

int
icp_marktq (int board)
{

  icp.osp[board]->clr_t_queue.sync = 0;
  icp.osp[board]->clr_t_queue.routine = icp_os_task_queue;
  icp.osp[board]->clr_t_queue.data = (void*)board;
  
  queue_task (&icp.osp[board]->clr_t_queue, &tq_scheduler);
  
  return OK;
}

/*
 * icp_irqoff
 */

void
icp_irqoff (int board)
{
  disable_irq (icp.pci[board]->irq_line);
}

/*
 * icp_irqon
 */

void
icp_irqon (int board)
{
  enable_irq (icp.pci[board]->irq_line);
}

/*
 * icp_splhigh
 */

void
icp_splhigh (int board)
{
  save_flags (icp.osp[board]->flags);
  cli();
}

/*
 * icp_spllow
 */

void
icp_spllow (int board)
{
  restore_flags (icp.osp[board]->flags);
}

/*
 * icp_fill_info
 */

int
icp_fill_info (int board, void* dummy_ptr)
{
  struct icp_info_t *frame_info = (struct icp_info_t*)dummy_ptr;

  frame_info->clock_tics = jiffies;
  do_gettimeofday (&(frame_info->time_stamp));
  memcpy (&(frame_info->frame_nums), &(icp.acc[board]),
	  sizeof (struct icp_acc_t));
  
  return OK;
}

/*
 * icp_assign_task
 */

int
icp_assign_task (int board)
{
  icp.osp[board]->task = current;

  return OK;
}

/*
 * icp_check_task
 */

int
icp_task_assigned (int board)
{
  return (icp.osp[board]->task != NULL);
}

/*
 * icp_deassign_task
 */

int
icp_deassign_task (int board)
{
  icp.osp[board]->task = NULL;

  return OK;
}

/*
 * icp_assign_signal
 */

int
icp_assign_signal (int board, icp_flags_t flags, icp_f_owner_t *f_owner)
{

  memcpy (&icp.osp[board]->fown, f_owner,
	  sizeof (struct fown_struct));
  
  /* enable signal generation if requested */
  if (flags & FASYNC)
    icp.osp[board]->fown.signum = SIGIO;
  else
    icp.osp[board]->fown.signum = 0;
  
  return OK;
}

/*
 * icp_deassign_signal
 */

int
icp_deassign_signal (int board)
{
  icp.osp[board]->fown.pid = 0;
  icp.osp[board]->fown.signum = 0;

  return OK;
}

/*
 * icp_check_blocked
 */

int
icp_check_blocked (int board)
{
  return ((icp.osp[board]->task->blocked.sig[0] >>
	   (ICP_SIGIO - 1)) & 1);
}

/*
 * icp_kill_proc
 */

#ifdef DEBUG_ICPCI
#  define DEBUG_ICPCI_IRQ
#endif

int
icp_kill_proc (int board)
{

  if (icp.osp[board]->fown.pid != 0 &&
      icp.osp[board]->fown.signum == ICP_SIGIO)
    {
#ifdef DEBUG_ICPCI_IRQ
      icp_irq_printf ("<s>");
#endif
      kill_proc (icp.osp[board]->fown.pid, ICP_SIGIO, 1);
    }
  
  return OK;
}

/*
 * icp_os_intr
 */

void
icp_os_intr (int irq, void *dev_id, struct pt_regs *regs)
{
  icp_intr (irq);
}

/*
 * icp_os_timeout
 */

void
icp_os_timeout (ulong data)
{
  icp_timeout ((int)data);
}

/*
 * icp_os_bottomhalf
 */

void
icp_os_task_queue (void *data)
{
  icp_task_queue ((int)data);
}

/*
 * icp_os_open
 */

static int
icp_os_open (struct inode *inode, struct file *file)
{
  int error;
  
  struct icp_devno_t devno;
  
  u_long flags;
  
  devno.major = MAJOR(inode->i_rdev);
  devno.major0 = major0;	/* dynamic (or explicit) assigned */
  devno.major1 = major1;	/* dynamic (or explicit) assigned */
  devno.minor = MINOR(inode->i_rdev);
  
  flags = file->f_flags;
  
  if ((error = icp_decode_devno (&devno)))
    return error;
  
  if ((error = icp_open (&devno, flags)))
    return error;
  
#ifdef MODULE    
  MOD_INC_USE_COUNT;
#endif    
  
  return OK;
}

/*
 * icp_os_release
 */

static int
icp_os_release (struct inode *inode, struct file *file)
{
  int error;
  
  struct icp_devno_t devno;
  
  u_long flags;
  
  devno.major = MAJOR(inode->i_rdev);
  devno.major0 = major0;	/* dynamic (or explicit) assigned */
  devno.major1 = major1;	/* dynamic (or explicit) assigned */
  devno.minor = MINOR(inode->i_rdev);
  
  flags = file->f_flags;
  
  if ((error = icp_decode_devno (&devno)))
    return error;
  
  if ((error = icp_close (&devno, flags)))
    return error;
  
#ifdef MODULE    
  MOD_DEC_USE_COUNT;
#endif

  return OK;
}

/*
 * icp_read
 */

static ssize_t
icp_os_read (struct file* file, char *buffer, size_t count, loff_t *offset)
{
  int error;

  struct icp_devno_t devno;

  devno.major = MAJOR(file->f_dentry->d_inode->i_rdev);
  devno.major0 = major0;	/* dynamic (or explicit) assigned */
  devno.major1 = major1;	/* dynamic (or explicit) assigned */
  devno.minor = MINOR(file->f_dentry->d_inode->i_rdev);

  if ((error = icp_decode_devno (&devno)))
    return error;
  
  return icp_read (&devno, file->f_flags, buffer, count);
}

/*
 * icp_os_write
 */

static int
icp_os_write (struct file *file, const char *buffer,
   	      size_t count, loff_t *offset)
{
  int error;

  struct icp_devno_t devno;

  devno.major = MAJOR(file->f_dentry->d_inode->i_rdev);
  devno.major0 = major0;	/* dynamic (or explicit) assigned */
  devno.major1 = major1;	/* dynamic (or explicit) assigned */
  devno.minor = MINOR(file->f_dentry->d_inode->i_rdev);

  if ((error = icp_decode_devno (&devno)))
    return error;
  
  return icp_write (&devno, file->f_flags, buffer, count);
}

/*
 * icp_os_mmap
 */

static int
icp_os_mmap (struct file *file, struct vm_area_struct *vma)
{
  int error;
  
  struct icp_devno_t devno;

  u_long base;
  size_t size;
  
  devno.major = MAJOR(file->f_dentry->d_inode->i_rdev);
  devno.major0 = major0;	/* dynamic (or explicit) assigned */
  devno.major1 = major1;	/* dynamic (or explicit) assigned */
  devno.minor = MINOR(file->f_dentry->d_inode->i_rdev);
  
  if ((error = icp_decode_devno (&devno)))
    return error;
  
  size = vma->vm_end - vma->vm_start;
  
  base = 0UL;
  if ((error = icp_mmap (&devno, &base, size)))
    return error;
  if (base == 0UL)
    return -EFAULT;

  /* remapping */
  if ((error = remap_page_range(vma->vm_start, base,
				size, vma->vm_page_prot)))
    return error;
  
  vma->vm_file = file;
  file->f_dentry->d_inode->i_count++;

  return OK;
}

/*
 * icp_os_select
 */

unsigned int
icp_os_poll (struct file *file, struct poll_table_struct *wait)
{
  int error, type, result, mask;
  
  struct icp_devno_t devno;
  
  devno.major = MAJOR(file->f_dentry->d_inode->i_rdev);
  devno.major0 = major0;	/* dynamic (or explicit) assigned */
  devno.major1 = major1;	/* dynamic (or explicit) assigned */
  devno.minor = MINOR(file->f_dentry->d_inode->i_rdev);
  
  if ((error = icp_decode_devno (&devno)))
    return error;

  ICP_PRINTF("icpci[%c]: icp_select - %s.\n", 'F', "TEST  IN|EX");
  ICP_PRINTF("icpci[%d]:       time : %ld.\n", devno.board, jiffies);
  
  mask = 0;
  /* check both in one */
  type = ICP_SEL_IN | ICP_SEL_EX;
  result = icp_select (&devno, type);
  
  if (result & ICP_SEL_IN)
    {
      ICP_PRINTF("icpci[%c]: icp_select - %s.\n", 'F', "FOUND IN");
      mask |= (POLLIN | POLLRDNORM);
    }
  else
    {
      ICP_PRINTF("icpci[%c]: icp_select - %s.\n", 'F', "WAIT  IN");
      poll_wait (file, &icp.osp[devno.board]->cam_w_queue, wait);
    }

  if (result & ICP_SEL_EX)
    {
      ICP_PRINTF("icpci[%c]: icp_select - %s.\n", 'F', "FOUND EX");
      mask |= POLLERR;
    }

  return mask;
}

/* 
 * icp_os_ioctl
 */

int
icp_os_ioctl (struct inode *inode, struct file *file,
	      u_int cmd, u_long arg)
{
  int error, count;
  
  struct icp_devno_t devno;

  union icp_inparam_t *inparam = NULL;
  union icp_outparam_t *outparam = NULL;
  
  devno.major = MAJOR(inode->i_rdev);
  devno.major0 = major0;	/* dynamic (or explicit) assigned */
  devno.major1 = major1;	/* dynamic (or explicit) assigned */
  devno.minor = MINOR(inode->i_rdev);

  if ((error = icp_decode_devno (&devno)))
    return error;
  
  /* validate module specific commands */
  if (_IOC_NR(cmd) > 31 && _IOC_NR(cmd) < 224)
    switch (icp.cap[devno.board]->module)
      {
      case ICP_AMFA:
	if (_IOC_NR(cmd) > 63 || _IOC_TYPE(cmd) != 'G')
	  return -EINVAL;
	break;
      case ICP_AMC1:
	if (_IOC_NR(cmd) < 64 || _IOC_NR(cmd) > 95 || _IOC_TYPE(cmd) != 'G')
	  return -EINVAL;
      break;
      case ICP_AMVS:
	if (_IOC_NR(cmd) < 96 || _IOC_NR(cmd) > 127 || _IOC_TYPE(cmd) != 'G')
	  return -EINVAL;
	break;
      case ICP_AMDG:
	if (_IOC_NR(cmd) < 128 || _IOC_NR(cmd) > 159 || _IOC_TYPE(cmd) != 'G')
	  return -EINVAL;
	break;
      case ICP_AMPV:
	if (_IOC_NR(cmd) < 96 || _IOC_NR(cmd) > 127 || _IOC_TYPE(cmd) != 'H')
	  return -EINVAL;
	break;
      default:
	return -ENXIO;
      }

  /* structure allocation */
  if (_IOC_DIR(cmd) == _IOC_READ ||
      _IOC_DIR(cmd) == (_IOC_READ | _IOC_WRITE))
    outparam = icp_malloc (sizeof(union icp_outparam_t), PAGED);

  if (_IOC_DIR(cmd) == _IOC_WRITE ||
      _IOC_DIR(cmd) == (_IOC_WRITE | _IOC_READ))
    inparam = icp_malloc (sizeof(union icp_inparam_t), PAGED);

  /* copyin */
  if (_IOC_DIR(cmd) == _IOC_WRITE ||
      _IOC_DIR(cmd) == (_IOC_WRITE | _IOC_READ))
    {
      if (cmd == GIOC_DG_SET_LUT)
	{
	  /* data has to be copied in single-page chunks */
	  for (count = 0; count < (sizeof(union dg_lut16_t) / PAGE_SIZE);
	       count++)
	    icp_copyin ((void *)inparam + count * PAGE_SIZE, 
			(void *)arg + count * PAGE_SIZE, PAGE_SIZE);
	  /* if the structure is not a multiple of PAGE_SIZE */
	  if (sizeof(union dg_lut16_t) % PAGE_SIZE)
	    icp_copyin ((void *)inparam + count * PAGE_SIZE, 
			(void *)arg + count * PAGE_SIZE,
			sizeof(union dg_lut16_t) % PAGE_SIZE);
	}
      else
	{
	  icp_copyin (inparam, (void *)arg, _IOC_SIZE(cmd));
	}
    }

  if ((error = icp_ioctl (&devno, file->f_flags, &file->f_owner,
			  cmd, inparam, outparam)))
    {
      if (outparam != NULL)
	icp_free (outparam, PAGED);
      if (inparam != NULL)
	icp_free (inparam, PAGED);
      return error;
    }

  /* copyout */
  if (_IOC_DIR(cmd) == _IOC_READ ||
      _IOC_DIR(cmd) == (_IOC_WRITE | _IOC_READ))
    {
      if (cmd == GIOC_DG_GET_LUT)
	{
	  /* data has to be copied in single-page chunks */
	  for (count = 0; count < (sizeof(union dg_lut16_t) / PAGE_SIZE);
	       count++)
	    icp_copyout ((void *)outparam + count * PAGE_SIZE, 
			 (void *)arg + count * PAGE_SIZE, PAGE_SIZE);
	  /* if the structure is not a multiple of PAGE_SIZE */
	  if (sizeof(union dg_lut16_t) % PAGE_SIZE)
	    icp_copyout ((void *)outparam + count * PAGE_SIZE, 
			 (void *)arg + count * PAGE_SIZE,
			 sizeof(union dg_lut16_t) % PAGE_SIZE);
	}
      else
	{
	  icp_copyout ((void *)arg, outparam, _IOC_SIZE(cmd));
	}
    }

  /* free structures */
  if (outparam != NULL)
    icp_free (outparam, PAGED);
  if (inparam != NULL)
    icp_free (inparam, PAGED);

  return OK;
}

/*
 * icp_os_init
 */

int
icp_os_init (int board)
{

  ICP_PRINTF("icpci[%c]: icp_os_init.\n", 'F');

  /* init os layer */
  icp.osp[board] =
    (struct icp_osp_t*)icp_malloc (sizeof (struct icp_osp_t), UNPAGED);
  if (icp.osp[board] == NULL)
    return -ENOMEM;
  memset (icp.osp[board], 0x00, sizeof(struct icp_osp_t));
  
  /* timer init */
  init_timer (&icp.osp[board]->cam_timer);
  init_timer (&icp.osp[board]->iop_timer);

  return OK;
}

/*
 * icp_os_pci_init
 */

int
icp_os_pci_init (int *boards)
{
  int board;
  
  struct pci_dev
    *icp_dev = NULL;
  
  u_char icp_bus;
  u_int icp_devfn;
  
  ICP_PRINTF("icpci[%c]: icp_os_pci_init.\n", 'F');
  
  /* is a pci bios installed and supported */
  if (!pci_present())
    {
      icp_printf ("icpci[%c]: no 'pcibios' present.\n", 'E');
      return -ENXIO;
    }
  
  /* find all boards */
  board = 0;
  while ((icp_dev = pci_find_device (PCI_VENDOR_ID_IMAGINGTECH,
				     PCI_DEVICE_ID_IMAGINGTECH_ICPCI,
				     icp_dev)))
    {
      
      /* allocate structure */
      icp.pci[board] =
	(struct icp_pci_t*)icp_malloc (sizeof (struct icp_pci_t), UNPAGED);
      if (icp.pci[board] == NULL)
	return -ENOMEM;
      memset (icp.pci[board], 0x00, sizeof(struct icp_pci_t));
      
      /* base addresses of various regions */
      icp.pci[board]->address0 = icp_dev->base_address[0];
      icp.pci[board]->address1 = icp_dev->base_address[1];
      icp.pci[board]->address2 = icp_dev->base_address[2];
      icp.pci[board]->address3 = icp_dev->base_address[3];
      
      /* interrupt line */
      icp.pci[board]->irq_line = icp_dev->irq;
      if (icp.pci[board]->irq_line < 3 || icp.pci[board]->irq_line > 15 )
	icp_printf ("icpci[%d]: really use interrupt %d?\n", board,
		    icp.pci[board]->irq_line);
      
      /* assign identification for further calls */
      icp_bus = icp_dev->bus->number;
      icp_devfn = icp_dev->devfn;
      
      /* read revision id */
      if (pcibios_read_config_byte (icp_bus, icp_devfn,
				    PCI_REVISION_ID,
				    &icp.pci[board]->revision))
	{
	  icp_printf ("icpci[%d]: can't read revision id.\n", board);
	  return -EIO;
	}
      
      /* read revision level from status register (bit 7) */
      if (pcibios_read_config_word (icp_bus, icp_devfn,
				    PCI_STATUS,
				    &icp.pci[board]->status))
	{
	  icp_printf ("icpci[%d]: can't read status register.\n", board);
	  return -EIO;
	}
      
      /* enable bus master mode and adjust latency timer */
      pci_set_master (icp_dev);
      
      ICP_PRINTF("icpci[%d]: Bus       : %d.\n", board, icp_bus);
      ICP_PRINTF("icpci[%d]: Device    : %d.\n", board, PCI_SLOT(icp_devfn));
      ICP_PRINTF("icpci[%d]: Function  : %d.\n", board, PCI_FUNC(icp_devfn));
      
      board++;
    }
  *boards = board;
  
  /* FIXME: for linux we use reserved upper memory */
  /* for the bigphysarea-patch we have to modify the kernel */
  /* the best way is to do a vmalloc_continuous (not available) */

  /* remap the reserved upper memory space (if available) */
  if (icp.bma_on && *boards)
    {
      /* get buffer begin and size in two different ways */
#ifdef ICP_AUTODETECT
      if ((icp.bma_size = icpci_memory_size) == 0UL)
	{
	  icp_printf ("icpci[%c]: no unused upper memory, "
		      "bma disabled!\n", 'W');
	  icp.bma_on = FALSE;
	  return OK;
	}
      else
	icp.p_bma_base = bios_memory_end - icpci_memory_size;
#else
      if (size)
	icp.bma_size = size * 1024 * 1024;
      else
	{
	  icp_printf ("icpci[%c]: no unused upper memory, "
		      "bma disabled!\n", 'W');
	  icp.bma_on = FALSE;
	  return OK;
	}
      if (base > 4)
	icp.p_bma_base = base * 1024 * 1024;
      else
	{
	  icp_printf ("icpci[%c]: no or wrong base address, "
		      "bma disabled!\n", 'W');
	  icp.bma_on = FALSE;
	  return OK;
	}
#endif
  
      /* begin and size are always aligned (multilple of 1Mb) */
      if((icp.v_bma_base = ioremap (icp.p_bma_base, icp.bma_size)) == NULL)
	{
	  icp_printf("icpci[%d]: unable to vremap unused upper memory.\n",
		     board);
	  return -EFAULT;
	}
    }

  return OK;
}

/*
 * icp_os_map_init
 */

int
icp_os_map_init (int board, u_long *align_div, ptrdiff_t *align_mod)
{

  ICP_PRINTF("icpci[%c]: icp_os_map_init.\n", 'F');

  /* remap all register spaces to low kernel memory */
  if (!icp.map[board]->regs_io)
    {
      icp.map[board]->v_icr_base =
	ioremap (align_div[ICR_REGS],
		 ROUND_UPTO_PAGE(align_mod[ICR_REGS] +
				 ICR_REG_SIZE * sizeof(u_long)));
      if (icp.map[board]->v_icr_base == NULL)
	{
	  icp_printf ("icpci[%d]: unable to ioremap interface control regs.\n",
		      board);
	  return -EFAULT;
	}
      icp.map[board]->v_cr_base =
	ioremap (align_div[CR_REGS],
		 ROUND_UPTO_PAGE(align_mod[CR_REGS] +
				 CR_REG_SIZE * sizeof(u_short)));
      if (icp.map[board]->v_cr_base == NULL)
	{
	  icp_printf ("icpci[%d]: unable to ioremap control regs.\n", board);
	  return -EFAULT;
	}
    }
  
  icp.map[board]->v_mod_base =
    ioremap (icp.pci[board]->address2, ROUND_UPTO_PAGE(MOD_MAP_SIZE));
  if (icp.map[board]->v_mod_base == NULL)
    {
    icp_printf ("icpci[%d]: unable to ioremap acquisition modul regs.\n",
		board);
    return -EFAULT;
    }
  
  if (icp.map[board]->big_mem)
    icp.map[board]->v_mem_base =
      ioremap (icp.pci[board]->address3, ROUND_UPTO_PAGE(MEM_MAP_SIZE << 1));
  else
    icp.map[board]->v_mem_base =
      ioremap (icp.pci[board]->address3, ROUND_UPTO_PAGE(MEM_MAP_SIZE));

  if(icp.map[board]->v_mem_base == NULL)
    {
      icp_printf ("icpci[%d]: unable to ioremap main board memory.\n",
		  board);
      return -EFAULT;
    }

  return OK;
}

/*
 * icp_os_cap_init
 */

int
icp_os_cap_init (int board)
{

  icp.cap[board]->io_wait8 = icp.cap[board]->mem_wait8 = wait8;
  icp.cap[board]->io_wait16 = icp.cap[board]->mem_wait16 = wait16;
  icp.cap[board]->io_wait32 = icp.cap[board]->mem_wait32 = wait32;

  return OK;
}

#ifdef MODULE
/*
 * cleanup_module
 */

void
cleanup_module (void)
{  
  int	intr, board, camera;
  
  ICP_PRINTF("icpci[%c]: cleanup_module.\n", 'F');
  
  for (board = 0; board < ICP_BOARDS_MAX; board++ )
    {
      if (icp.bma[board] != NULL)
	{
	  ICP_PRINTF("icp.bma[%d]: 0x%p\n", board, icp.bma[board]);
	  icp_free (icp.bma[board], UNPAGED);
	}
      if (icp.mem[board] != NULL)
	{
	  ICP_PRINTF("icp.mem[%d]: 0x%p\n", board, icp.mem[board]);
	  icp_free (icp.mem[board], UNPAGED);
	}
      for (camera = 0; camera < ICP_CAMERAS_MAX; camera++)
	if (icp.dev[board][camera] != NULL)
	  {
	    ICP_PRINTF("icp.dev[%d][%d]: 0x%p\n", board, camera,
		   icp.dev[board][camera]);
	    icp_free (icp.dev[board][camera], UNPAGED);
	  }
      if (icp.cap[board] != NULL)
	{
	  ICP_PRINTF("icp.cap[%d]: 0x%p\n", board, icp.cap[board]);
	  icp_free (icp.cap[board], UNPAGED);
	}
      if (icp.map[board] != NULL)
	{
	  if (icp.map[board]->v_mem_base != NULL)
	    {
	      ICP_PRINTF("icp.map[%d]->v_mem_base: 0x%p\n", board,
		     icp.map[board]->v_mem_base);
	      iounmap (icp.map[board]->v_mem_base);
	    }
	  if (icp.map[board]->v_mod_base != NULL)
	    {
	      ICP_PRINTF("icp.map[%d]->v_mod_base: 0x%p\n", board,
		     icp.map[board]->v_mod_base);
	      iounmap (icp.map[board]->v_mod_base);
	    }
	  if (icp.map[board]->v_cr_base != NULL)
	    {
	      icp.map[board]->v_cr_base =
		(u_short*)ROUND_DOWNTO_PAGE(icp.map[board]->v_cr_base);
	      ICP_PRINTF("icp.map[%d]->v_cr_base: 0x%p\n", board,
		     icp.map[board]->v_cr_base);
	      iounmap (icp.map[board]->v_cr_base);
	    }
	  if (icp.map[board]->v_icr_base != NULL)
	    {
	      icp.map[board]->v_icr_base =
		(u_long*)ROUND_DOWNTO_PAGE(icp.map[board]->v_icr_base);
	      ICP_PRINTF("icp.map[%d]->v_icr_base: 0x%p\n", board,
		     icp.map[board]->v_icr_base);
	      iounmap (icp.map[board]->v_icr_base);
	    }
	  ICP_PRINTF("icp.map[%d]: 0x%p\n", board, icp.map[board]);
	  icp_free (icp.map[board], UNPAGED);
	}
      if (icp.osp[board] != NULL)
	{
	  ICP_PRINTF("icp.osp[%d]: 0x%p\n", board,
		     icp.osp[board]);
	  icp_free (icp.osp[board], UNPAGED);
	}
      if (icp.pci[board] != NULL) {
	ICP_PRINTF("icp.pci[%d]: 0x%p\n", board, icp.pci[board]);
	icp_free (icp.pci[board], UNPAGED);
      }
    }
  if (icp.v_bma_base != NULL)
    {
      ICP_PRINTF("icp.v_bma_base[%d]: 0x%p\n", board, icp.v_bma_base);
      iounmap (icp.v_bma_base);
    }

  for (intr = 0; intr < ICP_NUM_X86INTS; intr++)
    if (used_intr & (1 << intr))
      {
	free_irq (intr, &icp);
	used_intr &= ~(1 << intr);
      }

  if (unregister_chrdev(major1, "icpciB") < 0)
    icp_printf ("icpci[%c]: error while unregistering 'icpciB' (%d).\n",
		'E', major1);
  if (unregister_chrdev(major0, "icpciA") < 0)
    icp_printf ("icpci[%c]: error while unregistering 'icpciA' (%d).\n",
		'E', major0);
}

/*
 * init_module
 */

int
init_module (void)
{
  int	error, board, boards;
  
  ICP_PRINTF("icpci[%c]: init_module.\n", 'F');
  
  /* how are the major devno's assigned (default is 0 - dynamic) */
  if (major0 != 0)
    {
      /* non automatic */
      if (major0 > MAX_CHRDEV)
	{
	  icp_printf ("icpci[%c]: major device number %d out of range "
		      " (1-%d).\n",
		      'I', major0, MAX_CHRDEV);
	  icp_printf ("icpci[%c]: using device number %d instead.\n",
		      'I', ICP_MAJOR_0);
	  major0 = ICP_MAJOR_0;	/* old driver behavior */
	}
      /* register first character device */
      if ((error = register_chrdev (major0, "icpciA", &icp_fops)) < 0)
	{
	  icp_printf ("icpci[%c]: 'register_chrdev (character device %d)' "
		      "failed.\n", 'I', major0);
	  cleanup_module();
	  return error;
	}
    }
  else
    {
      /* register first character device */
      if ((major0 = register_chrdev (major0, "icpciA", &icp_fops)) < 0)
	{
	  icp_printf ("icpci[%c]: 'register_chrdev (character device %d)' "
		      "failed.\n", 'I', major0);
	  cleanup_module();
	  return major0;
	}
    }  
  
  if (major1 != 0)
    {
      if (major1 > MAX_CHRDEV)
	{
	  /* non automatic */
	  icp_printf ("icpci[%c]: major device number %d out of range "
		      "(1-%d).\n",
		      'I', major1, MAX_CHRDEV);
	  icp_printf ("icpci[%c]: using device number %d instead.\n",
		      'I', ICP_MAJOR_1);
	  major1 = ICP_MAJOR_1;	/* old driver behavior */
	}
      /* register second character device */
      if ((error = register_chrdev (major1, "icpciB", &icp_fops)) < 0)
	{
	  icp_printf ("icpci[%c]: 'register_chrdev (character device %d)' "
		      "failed.\n", 'I', major1);
	  cleanup_module();
	  return error;
	}
    }
  else
    {
      /* register second character device */
      if ((major1 = register_chrdev (major1, "icpciB", &icp_fops)) < 0)
	{
	  icp_printf ("icpci[%c]: 'register_chrdev (character device %d)' "
		      "failed.\n", 'I', major1);
	  cleanup_module();
	  return major1;
	}
    }
  
  /* parse command line options */
  if (!strcmp (bma, "enable"))
    {
      icp.bma_on = TRUE;
      icp_printf ("icpci[%c]: bus master mode.\n", 'I');
    }
  else
    {
      icp.bma_on = FALSE;
      icp_printf ("icpci[%c]: normal memcpy mode.\n", 'I');	
    }
  
  /* check PCI - configuration space for possible devices */
  boards = 0;
  if ((error = icp_pci_init (&boards)) < 0)
    {
      cleanup_module();
      return error;
    }
  else
    if (boards == 0)
      {
	icp_printf ("icpci[%c]: no devices found.\n", 'I');
	{
	  cleanup_module();
	  return -ENXIO;
	}
      }
    else
      {
	icp_printf ("icpci[%c]: %d device(s) found.\n", 'I', boards);
	icp.boards = boards;
      }
  
  /* initialize data structures */
  for (board = 0; board < boards; board++)
    {    
      if ((error = icp_os_init (board)))
	{
	  cleanup_module();
	  return error;
	}
      if ((error = icp_map_init (board)))
	{
	  cleanup_module();
	  return error;
	}
      if ((error = icp_cap_init (board)))
	{
	  cleanup_module();
	  return error;
	}
      if ((error = icp_dev_init (board)))
	{
	  cleanup_module();
	  return error;
	}
    }
  
  for (board = 0; board < boards; board++)
    {
      if ((error = icp_mem_init (board)))
	{
	  cleanup_module();
	  return error;
	}
      if (icp.bma_on)
	if ((error = icp_bma_init (board)))
	  {
	    cleanup_module();
	    return error;
	  }
      /* reset statistics */
      icp.acc[board].frames_captured = 0UL;
      icp.acc[board].frames_valid = 0UL;
      icp.acc[board].frames_lost = 0UL;
      icp.acc[board].events_captured = 0UL;
      icp.acc[board].bytes_valid = 0UL;
      icp.acc[board].bytes_lost = 0UL;
      /* init on-board io-ports (newer boards only) */
      if (icp.pci[board]->revision > ICP_REV_A)
	if ((error = icp_iop_init (board)))
	  {
	    cleanup_module();
	    return error;
	  }
    }

  /* register per board interrupts */
  for (board = 0; board < boards; board++)
    if (!(used_intr & (1 << icp.pci[board]->irq_line)))
      {
	if ((error = request_irq (icp.pci[board]->irq_line,
				  &icp_os_intr,
				  SA_SHIRQ,
				  "icpci", &icp)))
	  if ((error = request_irq (icp.pci[board]->irq_line,
				    &icp_os_intr,
				    SA_INTERRUPT | SA_SHIRQ,
				    "icpci", &icp)))
	    {
	      icp_printf ("icpci[%c]: 'request_irq (intr. %d)' failed.\n",
			  'I', icp.pci[board]->irq_line);
	      cleanup_module();
		return error;
	    }
	used_intr |= (1 << icp.pci[board]->irq_line);
      }
  
  /* print version */
  icp_printf ("icpci[I]: %s\n", rcsid);

  return OK;
}
#endif /* MODULE */

